import sys
import argparse

from IPython.utils.process import arg_split
from IPython.core.magic import Magics, line_magic, magics_class
from IPython.core.magic_arguments import argument, magic_arguments
from sql import util
from sql.cmd.tables import tables
from sql.cmd.columns import columns
from sql.cmd.test import test
from sql.cmd.profile import profile
from sql.cmd.explore import explore
from sql.cmd.snippets import snippets

try:
    from traitlets.config.configurable import Configurable
except ModuleNotFoundError:
    from IPython.config.configurable import Configurable
from sql import exceptions


class CmdParser(argparse.ArgumentParser):
    def exit(self, status=0, message=None):
        if message:
            self._print_message(message, sys.stderr)

    def error(self, message):
        raise exceptions.UsageError(message)


@magics_class
class SqlCmdMagic(Magics, Configurable):
    """%sqlcmd magic"""

    @line_magic("sqlcmd")
    @magic_arguments()
    @argument("line", type=str, help="Command name")
    def _validate_execute_inputs(self, line):
        """
        Function to validate %sqlcmd inputs.
        Raises UsageError in case of an invalid input, executes command otherwise.
        """

        # We rely on SQLAlchemy when inspecting tables
        util.support_only_sql_alchemy_connection("%sqlcmd")

        AVAILABLE_SQLCMD_COMMANDS = [
            "tables",
            "columns",
            "test",
            "profile",
            "explore",
            "snippets",
        ]

        VALID_COMMANDS_MSG = (
            f"Missing argument for %sqlcmd. "
            f"Valid commands are: {', '.join(AVAILABLE_SQLCMD_COMMANDS)}"
        )

        if line == "":
            raise exceptions.UsageError(VALID_COMMANDS_MSG)
        else:
            split = arg_split(line)
            command, others = split[0].strip(), split[1:]

            if command in AVAILABLE_SQLCMD_COMMANDS:
                return self.execute(command, others)
            else:
                raise exceptions.UsageError(
                    f"%sqlcmd has no command: {command!r}. "
                    "Valid commands are: {}".format(
                        ", ".join(AVAILABLE_SQLCMD_COMMANDS)
                    )
                )

    @argument("cmd_name", default="", type=str, help="Command name")
    @argument("others", default="", type=str, help="Other tags")
    def execute(self, cmd_name="", others="", cell="", local_ns=None):
        """
        Command
        """

        router = {
            "tables": tables,
            "columns": columns,
            "test": test,
            "profile": profile,
            "explore": explore,
            "snippets": snippets,
        }

        cmd = router.get(cmd_name)
        if cmd:
            return cmd(others)



from jinja2 import Template
from sqlalchemy.exc import ProgrammingError

import sql.connection
from sql.util import flatten
from sql import exceptions


def _summary_stats(conn, table, column, with_=None):
    if conn.dialect in {"duckdb", "postgresql"}:
        return _summary_stats_parallel(conn, table, column, with_=with_)
    elif conn.dialect in {"redshift"}:
        return _summary_stats_redshift(conn, table, column, with_=with_)
    else:
        return _summary_stats_one_by_one(conn, table, column, with_=with_)


def _summary_stats_one_by_one(conn, table, column, with_=None):
    if not conn:
        conn = sql.connection.ConnectionManager.current.connection

    template_percentile = Template(
        """
SELECT
percentile_disc(0.25) WITHIN GROUP (ORDER BY "{{column}}") OVER (),
percentile_disc(0.50) WITHIN GROUP (ORDER BY "{{column}}") OVER (),
percentile_disc(0.75) WITHIN GROUP (ORDER BY "{{column}}") OVER ()
FROM "{{table}}"
"""
    )
    query = template_percentile.render(table=table, column=column)

    percentiles = list(conn.execute(query, with_).fetchone())

    template = Template(
        """
SELECT
AVG("{{column}}") AS mean,
COUNT(*) AS N
FROM "{{table}}"
"""
    )
    query = template.render(table=table, column=column)

    other = list(conn.execute(query, with_).fetchone())

    keys = ["q1", "med", "q3", "mean", "N"]

    return {k: float(v) for k, v in zip(keys, percentiles + other)}


def _summary_stats_redshift(conn, table, column, with_=None):
    if not conn:
        conn = sql.connection.ConnectionManager.current.connection

    template_percentile = Template(
        """
SELECT
approximate percentile_disc(0.25) WITHIN GROUP (ORDER BY "{{column}}"),
approximate percentile_disc(0.50) WITHIN GROUP (ORDER BY "{{column}}"),
approximate percentile_disc(0.75) WITHIN GROUP (ORDER BY "{{column}}")
FROM "{{table}}"
"""
    )
    query = template_percentile.render(table=table, column=column)

    percentiles = list(conn.execute(query, with_).fetchone())

    template = Template(
        """
SELECT
AVG("{{column}}") AS mean,
COUNT(*) AS N
FROM "{{table}}"
"""
    )
    query = template.render(table=table, column=column)

    other = list(conn.execute(query, with_).fetchone())

    keys = ["q1", "med", "q3", "mean", "N"]

    return {k: float(v) for k, v in zip(keys, percentiles + other)}


def _summary_stats_parallel(conn, table, column, with_=None):
    """Compute percentiles and mean for boxplot"""

    if not conn:
        conn = sql.connection.ConnectionManager.current

    driver = conn._get_database_information()["driver"]

    template = Template(
        """
    SELECT
    percentile_disc([0.25, 0.50, 0.75]) WITHIN GROUP \
    (ORDER BY "{{column}}") AS percentiles,
    AVG("{{column}}") AS mean,
    COUNT(*) AS N
    FROM "{{table}}"
"""
    )

    query = template.render(table=table, column=column)

    try:
        values = conn.execute(query, with_).fetchone()
    except ProgrammingError as e:
        print(e)
        raise exceptions.RuntimeError(
            f"\nEnsure that percentile_disc function is available on {driver}."
        )
    except Exception as e:
        raise e

    keys = ["q1", "med", "q3", "mean", "N"]
    return {k: float(v) for k, v in zip(keys, flatten(values))}



from typing import Iterator, Iterable
from collections.abc import MutableMapping
from jinja2 import Template
from ploomber_core.exceptions import modify_exceptions
import sql.connection
import difflib

from sql import exceptions
from sql import query_util


class SQLStore(MutableMapping):
    """Stores SQL scripts to render large queries with CTEs

    Notes
    -----
    .. versionadded:: 0.4.3

    Examples
    --------
    >>> from sql.store import SQLStore
    >>> sqlstore = SQLStore()
    >>> sqlstore.store("writers_fav",
    ...                "SELECT * FROM writers WHERE genre = 'non-fiction'")
    >>> sqlstore.store("writers_fav_modern",
    ...                "SELECT * FROM writers_fav WHERE born >= 1970",
    ...                with_=["writers_fav"])
    >>> query = sqlstore.render("SELECT * FROM writers_fav_modern LIMIT 10",
    ...                         with_=["writers_fav_modern"])
    >>> print(query)
    WITH "writers_fav" AS (
        SELECT * FROM writers WHERE genre = 'non-fiction'
    ), "writers_fav_modern" AS (
        SELECT * FROM writers_fav WHERE born >= 1970
    )
    SELECT * FROM writers_fav_modern LIMIT 10
    """

    def __init__(self):
        self._data = dict()

    def __setitem__(self, key: str, value: str) -> None:
        self._data[key] = value

    def __getitem__(self, key) -> str:
        if key not in self._data:
            matches = difflib.get_close_matches(key, self._data)
            error = f'"{key}" is not a valid snippet identifier.'
            if matches:
                raise exceptions.UsageError(error + f' Did you mean "{matches[0]}"?')
            else:
                valid = ", ".join(f'"{key}"' for key in self._data.keys())
                raise exceptions.UsageError(error + f" Valid identifiers are {valid}.")
        return self._data[key]

    def __iter__(self) -> Iterator[str]:
        for key in self._data:
            yield key

    def __len__(self) -> int:
        return len(self._data)

    def __delitem__(self, key: str) -> None:
        del self._data[key]

    def render(self, query, with_=None):
        # TODO: if with is false, WITH should not appear
        return SQLQuery(self, query, with_)

    def infer_dependencies(self, query, key):
        dependencies = []
        saved_keys = [
            saved_key for saved_key in list(self._data.keys()) if saved_key != key
        ]
        if saved_keys and query:
            tables = query_util.extract_tables_from_query(query)
            for table in tables:
                if table in saved_keys:
                    dependencies.append(table)
        return dependencies

    @modify_exceptions
    def store(self, key, query, with_=None):
        if "-" in key:
            raise exceptions.UsageError(
                "Using hyphens (-) in save argument isn't allowed."
                " Please use underscores (_) instead"
            )
        if with_ and key in with_:
            raise exceptions.UsageError(
                f"Script name ({key!r}) cannot appear in with_ argument"
            )

        self._data[key] = SQLQuery(self, query, with_)


class SQLQuery:
    """Holds queries and renders them"""

    def __init__(self, store: SQLStore, query: str, with_: Iterable = None):
        self._store = store
        self._query = query
        self._with_ = with_ or []

        if any("-" in x for x in self._with_):
            raise exceptions.UsageError(
                "Using hyphens is not allowed. "
                "Please use "
                + ", ".join(self._with_).replace("-", "_")
                + " instead for the with argument.",
            )

    def __str__(self) -> str:
        """
        We use the ' (backtick symbol) to wrap the CTE alias if the dialect supports
        ` (backtick)
        """
        with_clause_template = Template(
            """WITH{% for name in with_ %} {{name}} AS ({{rts(saved[name]._query)}})\
{{ "," if not loop.last }}{% endfor %}{{query}}"""
        )
        with_clause_template_backtick = Template(
            """WITH{% for name in with_ %} `{{name}}` AS ({{rts(saved[name]._query)}})\
{{ "," if not loop.last }}{% endfor %}{{query}}"""
        )
        is_use_backtick = (
            sql.connection.ConnectionManager.current.is_use_backtick_template()
        )
        with_all = _get_dependencies(self._store, self._with_)
        template = (
            with_clause_template_backtick if is_use_backtick else with_clause_template
        )
        # return query without 'with' when no dependency exists
        if len(with_all) == 0:
            return self._query.strip()
        return template.render(
            query=self._query,
            saved=self._store._data,
            with_=with_all,
            rts=_remove_trailing_semicolon,
        )


def _remove_trailing_semicolon(query):
    query_ = query.rstrip()
    return query_[:-1] if query_[-1] == ";" else query


def _get_dependencies(store, keys):
    """Get a list of all dependencies to reconstruct the CTEs in keys"""
    # get the dependencies for each key
    deps = _flatten([_get_dependencies_for_key(store, key) for key in keys])
    # remove duplicates but preserve order
    return list(dict.fromkeys(deps + keys))


def _get_dependents_for_key(store, key):
    key_dependents = []
    for k in list(store):
        deps = _get_dependencies_for_key(store, k)
        if key in deps:
            key_dependents.append(k)
    return key_dependents


def _get_dependencies_for_key(store, key):
    """Retrieve dependencies for a single key"""
    deps = store[key]._with_
    deps_of_deps = _flatten([_get_dependencies_for_key(store, dep) for dep in deps])
    return deps_of_deps + deps


def _flatten(elements):
    """Flatten a list of lists"""
    return [element for sub in elements for element in sub]


# session-wide store
store = SQLStore()



ORIGINAL_ERROR = "\nOriginal error message from DB driver:\n"
CTE_MSG = (
    "If using snippets, you may pass the --with argument explicitly.\n"
    "For more details please refer: "
    "https://jupysql.ploomber.io/en/latest/compose.html#with-argument"
)


def _is_syntax_error(error):
    """
    Function to detect whether error message from DB driver
    is related to syntax error in user query.
    """
    error_lower = error.lower()
    return (
        "syntax error" in error_lower
        or ("catalog error" in error_lower and "does not exist" in error_lower)
        or "error in your sql syntax" in error_lower
        or "incorrect syntax" in error_lower
        or "not found" in error_lower
    )


def detail(original_error):
    original_error = str(original_error)
    if _is_syntax_error(original_error):
        return f"{CTE_MSG}\n\n{ORIGINAL_ERROR}{original_error}\n"

    if "fe_sendauth: no password supplied" in original_error:
        specific_error = """\nLooks like you have run into some issues.
            Review our DB connection via URL strings guide:
            https://jupysql.ploomber.io/en/latest/connecting.html .
             Using Ubuntu? Check out this guide: "
            https://help.ubuntu.com/community/PostgreSQL#fe_sendauth:_
            no_password_supplied\n"""

        return f"{specific_error}\n{ORIGINAL_ERROR}{original_error}\n"

    return None



import re
import itertools
import shlex
from os.path import expandvars

from six.moves import configparser as CP
from sqlalchemy.engine.url import URL
from IPython.core.magic_arguments import parse_argstring


def connection_from_dsn_section(section, config):
    parser = CP.ConfigParser()
    parser.read(config.dsn_filename)
    cfg_dict = dict(parser.items(section))
    return str(URL.create(**cfg_dict).render_as_string(hide_password=False))


def _connection_string(s, config):
    s = expandvars(s)  # for environment variables
    if "@" in s or "://" in s:
        return s
    if s.startswith("[") and s.endswith("]"):
        section = s.lstrip("[").rstrip("]")
        parser = CP.ConfigParser()
        parser.read(config.dsn_filename)
        cfg_dict = dict(parser.items(section))
        return str(URL.create(**cfg_dict))
    return ""


def parse(cell, config):
    """Extract connection info and result variable from SQL

    Please don't add any more syntax requiring
    special parsing.
    Instead, add @arguments to SqlMagic.execute.

    We're grandfathering the
    connection string and `<<` operator in.
    """
    result = {
        "connection": "",
        "sql": "",
        "result_var": None,
        "return_result_var": False,
    }

    pieces = cell.split(None, 1)
    if not pieces:
        return result
    result["connection"] = _connection_string(pieces[0], config)
    if result["connection"]:
        if len(pieces) == 1:
            return result
        cell = pieces[1]

    pointer = cell.find("<<")
    if pointer != -1:
        left = cell[:pointer].replace(" ", "").replace("\n", "")
        right = cell[pointer + 2 :].strip(" ")

        if "=" in left:
            result["result_var"] = left[:-1]
            result["return_result_var"] = True
        else:
            result["result_var"] = left

        result["sql"] = right
    else:
        result["sql"] = cell
    return result


def _option_strings_from_parser(parser):
    """Extracts the expected option strings (-a, --append, etc) from argparse parser

    Thanks Martijn Pieters
    https://stackoverflow.com/questions/28881456/how-can-i-list-all-registered-arguments-from-an-argumentparser-instance

    :param parser: [description]
    :type parser: IPython.core.magic_arguments.MagicArgumentParser
    """
    opts = [a.option_strings for a in parser._actions]
    return list(itertools.chain.from_iterable(opts))


def without_sql_comment(parser, line):
    """Strips -- comment from a line

    The argparser unfortunately expects -- to precede an option,
    but in SQL that delineates a comment.  So this removes comments
    so a line can safely be fed to the argparser.

    :param line: A line of SQL, possibly mixed with option strings
    :type line: str
    """

    args = _option_strings_from_parser(parser)
    result = itertools.takewhile(
        lambda word: (not word.startswith("--")) or (word in args),
        shlex.split(line, posix=False),
    )
    return " ".join(result)


def magic_args(magic_execute, line):
    line = without_sql_comment(parser=magic_execute.parser, line=line)
    return parse_argstring(magic_execute, line)


def escape_string_literals_with_colon_prefix(query):
    """
    Given a query, replaces all occurrences of ':variable' with '\:variable' and
    ":variable" with "\:variable" so that the query can be passed to sqlalchemy.text
    without the literals being interpreted as bind parameters. It doesn't replace
    the occurrences of :variable (without quotes)
    """  # noqa

    # Define the regular expression pattern for valid Python identifiers
    identifier_pattern = r"\b[a-zA-Z_][a-zA-Z0-9_]*\b"

    double_quoted_variable_pattern = r'(?<!\\)":(' + identifier_pattern + r')(?<!\\)"'

    # Define the regular expression pattern for matching ':variable' format
    single_quoted_variable_pattern = r"(?<!\\)':(" + identifier_pattern + r")(?<!\\)\'"

    # Replace ":variable" and ':variable' with "\:variable"
    query_quoted = re.sub(double_quoted_variable_pattern, r'"\\:\1"', query)
    query_quoted = re.sub(single_quoted_variable_pattern, r"'\\:\1'", query_quoted)

    double_found = re.findall(double_quoted_variable_pattern, query)
    single_found = re.findall(single_quoted_variable_pattern, query)

    return query_quoted, double_found + single_found


def find_named_parameters(input_string):
    # Define the regular expression pattern for valid Python identifiers
    identifier_pattern = r"\b[a-zA-Z_][a-zA-Z0-9_]*\b"

    # Define the regular expression pattern for matching :variable format
    variable_pattern = r'(?<!["\'])\:(' + identifier_pattern + ")"

    # Use findall to extract all matches of :variable from the input string
    matches = re.findall(variable_pattern, input_string)

    return matches



from abc import abstractmethod


class geom:
    """
    Base class of all geom
    """

    def __init__(self):
        pass

    def __add__(self, gg):
        return gg

    def __radd__(self, gg):
        return gg + self

    @abstractmethod
    def draw(self, gg):
        """
        Draws plot
        """
        pass



from sql import inspect
from sql.cmd.cmd_utils import CmdParser


def profile(others):
    """
    Implementation of `%sqlcmd profile`
    This function takes in a string containing command line arguments,
    parses them to extract the name of the table, the schema, and the output location.
    It then retrieves statistical information about the specified table and either
    returns the report or writes it to the specified location.


    Parameters
    ----------
    others : str,
        A string containing the command line arguments.

    Returns
    -------
    report: PrettyTable
        statistics of the table
    """
    parser = CmdParser()
    parser.add_argument("-t", "--table", type=str, help="Table name", required=True)

    parser.add_argument("-s", "--schema", type=str, help="Schema name", required=False)

    parser.add_argument(
        "-o", "--output", type=str, help="Store report location", required=False
    )

    args = parser.parse_args(others)

    report = inspect.get_table_statistics(schema=args.schema, name=args.table)

    if args.output:
        with open(args.output, "w") as f:
            f.write(report._repr_html_())

    return report



from sql.telemetry import telemetry


class aes:
    """
    Aesthetic mappings

    Parameters
    ----------
    x: str | list
        x aesthetic mapping

    fill : str
        The inner color of a shape

    color : str, default 'None'
        The edge color of a shape
    """

    @telemetry.log_call("aes-init")
    def __init__(self, x=None, fill=None, color=None):
        self.x = x
        self.fill = fill
        self.color = color



from sqlglot import parse_one, exp
from sqlglot.errors import ParseError


def extract_tables_from_query(query):
    """
    Function to extract names of tables from
    a syntactically correct query

    Parameters
    ----------
    query : str, user query

    Returns
    -------
    list
        List of tables in the query
        [] if error in parsing the query
    """
    try:
        tables = [table.name for table in parse_one(query).find_all(exp.Table)]
        return tables
    except ParseError:
        # TODO : Instead of returning [] replace with call to
        # error_messages.py::parse_sqlglot_error. Currently this
        # is not possible because of an exception raised in test
        # fixtures. (#546). This function can also be moved to util.py
        # after #545 is resolved.
        return []



from sql.widgets import TableWidget
from IPython.display import display
from sql.cmd.cmd_utils import CmdParser


def explore(others):
    """
    Implementation of `%sqlcmd explore`
    This function takes in a string containing command line arguments,
    parses them to extract the name of the table, and displays an interactive
    widget for exploring the contents of the specified table.

    Parameters
    ----------
    others : str,
        A string containing the command line arguments.

    """
    parser = CmdParser()
    parser.add_argument("-t", "--table", type=str, help="Table name", required=True)
    args = parser.parse_args(others)

    table_widget = TableWidget(args.table)
    display(table_widget)



import os
from contextlib import contextmanager
import sys
import time

from sqlalchemy.engine import URL
import sqlalchemy
from IPython.core.interactiveshell import InteractiveShell

from ploomber_core.dependencies import requires

# SQLite and DuckDB do not require Docker, so we make docker packages optional
# in case we want to run those tests

try:
    from dockerctx import new_container
except ModuleNotFoundError:
    new_container = None

try:
    import docker
except ModuleNotFoundError:
    docker = None


TMP_DIR = "tmp"


class TestingShell(InteractiveShell):
    """
    A custom InteractiveShell that raises exceptions instead of silently suppressing
    them.
    """

    def run_cell(self, *args, **kwargs):
        result = super().run_cell(*args, **kwargs)
        result.raise_error()
        return result


class DatabaseConfigHelper:
    @staticmethod
    def get_database_config(database):
        return databaseConfig[database]

    @staticmethod
    def get_database_url(database):
        return _get_database_url(database)

    @staticmethod
    def get_tmp_dir():
        return TMP_DIR


mssql_base = {
    "username": "sa",
    "password": "Ploomber_App@Root_Password",
    "database": "master",
    "host": "localhost",
    "port": "1433",
    "query": {
        "driver": "ODBC Driver 18 for SQL Server",
        "Encrypt": "yes",
        "TrustServerCertificate": "yes",
    },
    "docker_ct": {
        "name": "MSSQL",
        "image": "mcr.microsoft.com/azure-sql-edge",
        "ports": {1433: 1433},
    },
    "alias": "MSSQLTest",
}

mssql_pyobdc = {**mssql_base, "drivername": "mssql+pyodbc"}
mssql_pytds = {**mssql_base, "drivername": "mssql+pytds"}

databaseConfig = {
    "postgreSQL": {
        "drivername": "postgresql",
        "username": "ploomber_app",
        "password": "ploomber_app_password",
        "database": "db",
        "host": "localhost",
        "port": "5432",
        "alias": "postgreSQLTest",
        "docker_ct": {
            "name": "postgres",
            "image": "postgres",
            "ports": {5432: 5432},
        },
        "query": {},
    },
    "mySQL": {
        "drivername": "mysql+pymysql",
        "username": "ploomber_app",
        "password": "ploomber_app_password",
        "root_password": "ploomber_app_root_password",
        "database": "db",
        "host": "localhost",
        "port": "33306",
        "alias": "mySQLTest",
        "docker_ct": {
            "name": "mysql",
            "image": "mysql",
            "ports": {3306: 33306},
        },
        "query": {},
    },
    "mariaDB": {
        "drivername": "mysql+pymysql",
        "username": "ploomber_app",
        "password": "ploomber_app_password",
        "root_password": "ploomber_app_root_password",
        "database": "db",
        "host": "localhost",
        "port": "33309",
        "alias": "mariaDBTest",
        "docker_ct": {
            "name": "mariadb",
            "image": "mariadb:10.4.30",
            "ports": {3306: 33309},
        },
        "query": {},
    },
    "SQLite": {
        "drivername": "sqlite",
        "username": None,
        "password": None,
        "database": "/{}/db-sqlite".format(TMP_DIR),
        "host": None,
        "port": None,
        "alias": "SQLiteTest",
        "query": {},
    },
    "duckDB": {
        "drivername": "duckdb",
        "username": None,
        "password": None,
        "database": "/{}/db-duckdb".format(TMP_DIR),
        "host": None,
        "port": None,
        "alias": "duckDBTest",
        "query": {},
    },
    "MSSQL": mssql_pyobdc,
    "mssql_pytds": mssql_pytds,
    "Snowflake": {
        "drivername": "snowflake",
        "username": os.getenv("SF_USERNAME"),
        "password": os.getenv("SF_PASSWORD"),
        # database/schema
        "database": os.getenv("SF_DATABASE", "JUPYSQL_INTEGRATION_TESTING/GENERAL"),
        "host": "lpb17716.us-east-1",
        "port": None,
        "alias": "snowflakeTest",
        "docker_ct": None,
        "query": {
            "warehouse": "COMPUTE_WH",
            "role": "SYSADMIN",
        },
    },
    "oracle": {
        "drivername": "oracle+oracledb",
        "username": "ploomber_app",
        "password": "ploomber_app_password",
        "admin_password": "ploomber_app_admin_password",
        # database/schema
        "host": "localhost",
        "port": "1521",
        "alias": "oracle",
        "database": None,
        "docker_ct": {
            "name": "oracle",
            "image": "gvenzl/oracle-xe",
            "ports": {1521: 1521},
        },
        "query": {
            "service_name": "XEPDB1",
        },
    },
    "redshift": {
        "drivername": "redshift+redshift_connector",
        "username": os.getenv("REDSHIFT_USERNAME"),
        "password": os.getenv("REDSHIFT_PASSWORD"),
        # database/schema
        "database": "dev",
        "host": os.getenv("REDSHIFT_HOST"),
        "port": 5439,
        "alias": "redshift",
        "docker_ct": None,
        "query": {},
    },
}


# SQLAlchmey URL: https://docs.sqlalchemy.org/en/20/core/engines.html#database-urls
def _get_database_url(database):
    return URL.create(
        drivername=databaseConfig[database]["drivername"],
        username=databaseConfig[database]["username"],
        password=databaseConfig[database]["password"],
        host=databaseConfig[database]["host"],
        port=databaseConfig[database]["port"],
        database=databaseConfig[database]["database"],
        query=databaseConfig[database]["query"],
    ).render_as_string(hide_password=False)


def database_ready(
    database,
    timeout=60,
    poll_freq=0.5,
):
    """Wait until the container is ready to receive connections.


    :type host: str
    :type port: int
    :type timeout: float
    :type poll_freq: float
    """
    errors = []

    t0 = time.time()
    while time.time() - t0 < timeout:
        try:
            eng = sqlalchemy.create_engine(_get_database_url(database)).connect()
            eng.close()
            print(f"{database} is initialized successfully")
            return True
        except ModuleNotFoundError:
            raise
        except Exception as e:
            print(type(e))
            errors.append(str(e))

        time.sleep(poll_freq)

    # print all the errors so we know what's going on since failing to connect might be
    # to some misconfiguration error
    errors_ = "\n".join(errors)
    print(f"ERRORS: {errors_}")

    return True


def get_docker_client():
    return docker.from_env(
        version="auto", environment={"DOCKER_HOST": os.getenv("DOCKER_HOST")}
    )


@contextmanager
@requires(["docker", "dockerctx"])
def postgres(is_bypass_init=False):
    if is_bypass_init:
        yield None
        return
    db_config = DatabaseConfigHelper.get_database_config("postgreSQL")
    try:
        client = get_docker_client()
        container = client.containers.get(db_config["docker_ct"]["name"])
        yield container
    except docker.errors.NotFound:
        print("Creating new container: postgreSQL")
        with new_container(
            new_container_name=db_config["docker_ct"]["name"],
            image_name=db_config["docker_ct"]["image"],
            ports=db_config["docker_ct"]["ports"],
            environment={
                "POSTGRES_DB": db_config["database"],
                "POSTGRES_USER": db_config["username"],
                "POSTGRES_PASSWORD": db_config["password"],
            },
            ready_test=lambda: database_ready(database="postgreSQL"),
            healthcheck={
                "test": "pg_isready",
                "interval": 10000000000,
                "timeout": 5000000000,
                "retries": 5,
            },
        ) as container:
            yield container


@contextmanager
@requires(["docker", "dockerctx"])
def mysql(is_bypass_init=False):
    if is_bypass_init:
        yield None
        return
    db_config = DatabaseConfigHelper.get_database_config("mySQL")
    try:
        client = get_docker_client()
        container = client.containers.get(db_config["docker_ct"]["name"])
        yield container
    except docker.errors.NotFound:
        print("Creating new container: mysql")
        with new_container(
            new_container_name=db_config["docker_ct"]["name"],
            image_name=db_config["docker_ct"]["image"],
            ports=db_config["docker_ct"]["ports"],
            environment={
                "MYSQL_DATABASE": db_config["database"],
                "MYSQL_USER": db_config["username"],
                "MYSQL_PASSWORD": db_config["password"],
                "MYSQL_ROOT_PASSWORD": db_config["root_password"],
            },
            command="mysqld --default-authentication-plugin=mysql_native_password",
            ready_test=lambda: database_ready(database="mySQL"),
            healthcheck={
                "test": [
                    "CMD",
                    "mysqladmin",
                    "ping",
                    "-h",
                    "localhost",
                    "--user=root",
                    "--password=ploomber_app_root_password",
                ],
                "timeout": 5000000000,
            },
        ) as container:
            yield container


@contextmanager
@requires(["docker", "dockerctx"])
def mariadb(is_bypass_init=False):
    if is_bypass_init:
        yield None
        return
    db_config = DatabaseConfigHelper.get_database_config("mariaDB")
    try:
        client = get_docker_client()
        curr = client.containers.get(db_config["docker_ct"]["name"])
        yield curr
    except docker.errors.NotFound:
        print("Creating new container: mariaDB")
        with new_container(
            new_container_name=db_config["docker_ct"]["name"],
            image_name=db_config["docker_ct"]["image"],
            ports=db_config["docker_ct"]["ports"],
            environment={
                "MYSQL_DATABASE": db_config["database"],
                "MYSQL_USER": db_config["username"],
                "MYSQL_PASSWORD": db_config["password"],
                "MYSQL_ROOT_PASSWORD": db_config["root_password"],
            },
            command="mysqld --default-authentication-plugin=mysql_native_password",
            ready_test=lambda: database_ready(database="mariaDB"),
            healthcheck={
                "test": [
                    "CMD",
                    "mysqladmin",
                    "ping",
                    "-h",
                    "localhost",
                    "--user=root",
                    "--password=ploomber_app_root_password",
                ],
                "timeout": 5000000000,
            },
        ) as container:
            yield container


@contextmanager
@requires(["docker", "dockerctx"])
def mssql(is_bypass_init=False):
    if is_bypass_init:
        yield None
        return
    db_config = DatabaseConfigHelper.get_database_config("MSSQL")
    try:
        client = get_docker_client()
        curr = client.containers.get(db_config["docker_ct"]["name"])
        yield curr
    except docker.errors.NotFound:
        print("Creating new container: MSSQL")
        with new_container(
            new_container_name=db_config["docker_ct"]["name"],
            image_name=db_config["docker_ct"]["image"],
            ports=db_config["docker_ct"]["ports"],
            environment={
                "MSSQL_DATABASE": db_config["database"],
                "MSSQL_USER": db_config["username"],
                "MSSQL_SA_PASSWORD": db_config["password"],
                "ACCEPT_EULA": "Y",
            },
            ready_test=lambda: database_ready(database="MSSQL"),
            healthcheck={
                "test": "/opt/mssql-tools/bin/sqlcmd "
                "-U $DB_USER -P $SA_PASSWORD "
                "-Q 'select 1' -b -o /dev/null",
                "timeout": 5000000000,
            },
        ) as container:
            yield container


@contextmanager
@requires(["docker", "dockerctx"])
def oracle(is_bypass_init=False):
    if is_bypass_init:
        yield None
        return
    db_config = DatabaseConfigHelper.get_database_config("oracle")
    try:
        client = get_docker_client()
        curr = client.containers.get(db_config["docker_ct"]["name"])
        yield curr
    except docker.errors.NotFound:
        print("Creating new container: oracle")
        with new_container(
            new_container_name=db_config["docker_ct"]["name"],
            image_name=db_config["docker_ct"]["image"],
            ports=db_config["docker_ct"]["ports"],
            environment={
                "APP_USER": db_config["username"],
                "APP_USER_PASSWORD": db_config["password"],
                "ORACLE_PASSWORD": db_config["admin_password"],
            },
            # Oracle takes more time to initialize
            ready_test=lambda: database_ready(database="oracle"),
        ) as container:
            yield container


def main():
    print("Starting test containers...")
    with postgres(), mysql(), mariadb(), mssql(), oracle():
        print("Press CTRL+C to exit")
        try:
            while True:
                time.sleep(5)
        except KeyboardInterrupt:
            print("Exit, containers will be killed")
            sys.exit()


if __name__ == "__main__":
    main()



from sql.magic import RenderMagic, SqlMagic, load_ipython_extension
from sql.connection import PLOOMBER_DOCS_LINK_STR

__version__ = "0.9.1"


__all__ = [
    "RenderMagic",
    "SqlMagic",
    "load_ipython_extension",
    "PLOOMBER_DOCS_LINK_STR",
]



from IPython.core.magic import (
    Magics,
    line_magic,
    magics_class,
)
from IPython.core.magic_arguments import argument, magic_arguments
from ploomber_core.exceptions import modify_exceptions

try:
    from traitlets.config.configurable import Configurable
except ModuleNotFoundError:
    from IPython.config.configurable import Configurable


from sql import plot
from sql.command import SQLPlotCommand
from sql import exceptions
from sql import util

SUPPORTED_PLOTS = ["histogram", "boxplot", "bar", "pie"]


@magics_class
class SqlPlotMagic(Magics, Configurable):
    """%sqlplot magic"""

    @line_magic("sqlplot")
    @magic_arguments()
    @argument("line", default="", nargs="*", type=str, help="Plot name")
    @argument("-t", "--table", type=str, help="Table to use", required=True)
    @argument(
        "-c", "--column", type=str, nargs="+", help="Column(s) to use", required=True
    )
    @argument(
        "-b",
        "--bins",
        type=int,
        default=50,
        help="Histogram bins",
    )
    @argument(
        "-o",
        "--orient",
        type=str,
        default="v",
        help="Boxplot orientation (v/h)",
    )
    @argument(
        "-w",
        "--with",
        type=str,
        help="Use a saved query",
        action="append",
        dest="with_",
    )
    @argument(
        "-S",
        "--show-numbers",
        action="store_true",
        help="Show number of observations",
    )
    @argument(
        "-B",
        "--breaks",
        type=float,
        nargs="+",
        help="Histogram breaks",
    )
    @modify_exceptions
    def execute(self, line="", cell="", local_ns=None):
        """
        Plot magic
        """

        cmd = SQLPlotCommand(self, line)

        if len(cmd.args.column) == 1:
            column = cmd.args.column[0]
        else:
            column = cmd.args.column

        if not cmd.args.line:
            plot_str = util.pretty_print(SUPPORTED_PLOTS, last_delimiter="or")
            raise exceptions.UsageError(
                "Missing the first argument, must be any of: "
                f"{plot_str}\n"
                "Example: %sqlplot histogram"
            )

        if cmd.args.line[0] not in SUPPORTED_PLOTS + ["hist", "box"]:
            plot_str = util.pretty_print(SUPPORTED_PLOTS, last_delimiter="or")
            raise exceptions.UsageError(
                f"Unknown plot {cmd.args.line[0]!r}. Must be any of: " f"{plot_str}"
            )

        column = util.sanitize_identifier(column)
        table = util.sanitize_identifier(cmd.args.table)

        if cmd.args.with_:
            with_ = cmd.args.with_
        else:
            with_ = self._check_table_exists(table)

        if cmd.args.line[0] in {"box", "boxplot"}:
            return plot.boxplot(
                table=table,
                column=column,
                with_=with_,
                orient=cmd.args.orient,
                conn=None,
            )
        elif cmd.args.line[0] in {"hist", "histogram"}:
            # to avoid passing bins default value when breaks are given by a user
            bin_specified = " --bins " in line or " -b " in line
            breaks_specified = " --breaks " in line or " -B " in line
            if breaks_specified and not bin_specified:
                bins = None
            else:
                bins = cmd.args.bins

            return plot.histogram(
                table=table,
                column=column,
                bins=bins,
                with_=with_,
                conn=None,
                breaks=cmd.args.breaks,
            )
        elif cmd.args.line[0] in {"bar"}:
            return plot.bar(
                table=table,
                column=column,
                with_=with_,
                orient=cmd.args.orient,
                show_num=cmd.args.show_numbers,
                conn=None,
            )
        elif cmd.args.line[0] in {"pie"}:
            return plot.pie(
                table=table,
                column=column,
                with_=with_,
                show_num=cmd.args.show_numbers,
                conn=None,
            )

    @staticmethod
    def _check_table_exists(table):
        with_ = None
        if util.is_saved_snippet(table):
            with_ = [table]
        else:
            util.is_table_exists(table)
        return with_



import warnings
from sql import inspect
import difflib
from sql.connection import ConnectionManager
from sql.store import store, _get_dependents_for_key
from sql import exceptions, display
import json
from pathlib import Path
from ploomber_core.dependencies import requires

try:
    import toml
except ModuleNotFoundError:
    toml = None

SINGLE_QUOTE = "'"
DOUBLE_QUOTE = '"'


def sanitize_identifier(identifier):
    if (identifier[0] == SINGLE_QUOTE and identifier[-1] == SINGLE_QUOTE) or (
        identifier[0] == DOUBLE_QUOTE and identifier[-1] == DOUBLE_QUOTE
    ):
        return identifier[1:-1]
    else:
        return identifier


def convert_to_scientific(value):
    """
    Converts value to scientific notation if necessary

    Parameters
    ----------
    value : any
        Value to format.
    """
    if (
        isinstance(value, (int, float))
        and not isinstance(value, bool)
        and _is_long_number(value)
    ):
        new_value = "{:,.3e}".format(value)

    else:
        new_value = value

    return new_value


def _is_long_number(num) -> bool:
    """
    Checks if num's digits > 10
    """
    if "." in str(num):
        split_by_decimal = str(num).split(".")
        if len(split_by_decimal[0]) > 10 or len(split_by_decimal[1]) > 10:
            return True
    return False


def get_suggestions_message(suggestions):
    suggestions_message = ""
    if len(suggestions) > 0:
        _suggestions_string = pretty_print(suggestions, last_delimiter="or")
        suggestions_message = f"\nDid you mean : {_suggestions_string}"
    return suggestions_message


def is_table_exists(
    table: str,
    schema: str = None,
    ignore_error: bool = False,
    conn=None,
) -> bool:
    """
    Checks if a given table exists for a given connection

    Parameters
    ----------
    table: str
        Table name

    schema: str, default None
        Schema name

    ignore_error: bool, default False
        Avoid raising a ValueError
    """
    if table is None:
        if ignore_error:
            return False
        else:
            raise exceptions.UsageError("Table cannot be None")
    if not ConnectionManager.current:
        raise exceptions.RuntimeError("No active connection")
    if not conn:
        conn = ConnectionManager.current

    table = strip_multiple_chars(table, "\"'")

    if schema:
        table_ = f"{schema}.{table}"
    else:
        table_ = table

    _is_exist = _is_table_exists(table_, conn)

    if not _is_exist:
        if not ignore_error:
            try_find_suggestions = not conn.is_dbapi_connection
            expected = []
            existing_schemas = []
            existing_tables = []

            if try_find_suggestions:
                existing_schemas = inspect.get_schema_names()

            if schema and schema not in existing_schemas:
                expected = existing_schemas
                invalid_input = schema
            else:
                if try_find_suggestions:
                    existing_tables = _get_list_of_existing_tables()

                expected = existing_tables
                invalid_input = table

            if schema:
                err_message = (
                    f"There is no table with name {table!r} in schema {schema!r}"
                )
            else:
                err_message = (
                    f"There is no table with name {table!r} in the default schema"
                )

            if table not in list(store):
                suggestions = difflib.get_close_matches(invalid_input, expected)
                suggestions_store = difflib.get_close_matches(
                    invalid_input, list(store)
                )
                suggestions.extend(suggestions_store)
                suggestions_message = get_suggestions_message(suggestions)
                if suggestions_message:
                    err_message = f"{err_message}{suggestions_message}"
            raise exceptions.TableNotFoundError(err_message)

    return _is_exist


def _get_list_of_existing_tables() -> list:
    """
    Returns a list of table names for a given connection
    """
    tables = []
    tables_rows = inspect.get_table_names()._table
    for row in tables_rows:
        table_name = row.get_string(fields=["Name"], border=False, header=False).strip()

        tables.append(table_name)
    return tables


def pretty_print(
    obj: list, delimiter: str = ",", last_delimiter: str = "and", repr_: bool = False
) -> str:
    """
    Returns a formatted string representation of an array
    """
    if repr_:
        sorted_ = sorted(repr(element) for element in obj)
    else:
        sorted_ = sorted(f"'{element}'" for element in obj)

    if len(sorted_) > 1:
        sorted_[-1] = f"{last_delimiter} {sorted_[-1]}"

    return f"{delimiter} ".join(sorted_)


def strip_multiple_chars(string: str, chars: str) -> str:
    """
    Trims characters from the start and end of the string
    """
    return string.translate(str.maketrans("", "", chars))


def is_saved_snippet(table: str) -> bool:
    if table in list(store):
        display.message(f"Plotting using saved snippet : {table}")
        return True
    return False


def _is_table_exists(table: str, conn) -> bool:
    """
    Runs a SQL query to check if table exists
    """
    if not conn:
        conn = ConnectionManager.current

    identifiers = conn.get_curr_identifiers()

    for iden in identifiers:
        if isinstance(iden, tuple):
            query = "SELECT * FROM {0}{1}{2} WHERE 1=0".format(iden[0], table, iden[1])
        else:
            query = "SELECT * FROM {0}{1}{0} WHERE 1=0".format(iden, table)
        try:
            conn.execute(query)
            return True
        except Exception:
            pass

    return False


def flatten(src, ltypes=(list, tuple)):
    """The flatten function creates a new tuple / list
    with all sub-tuple / sub-list elements concatenated into it recursively

    Parameters
    ----------
    src : tuple / list
        Source tuple / list with all sub-tuple / sub-list elements
    ltypes : tuple, optional
        sub element's data type, by default (list, tuple)

    Returns
    -------
    tuple / list
        Flatten tuple / list
    """
    ltype = type(src)
    # Create a process list to handle flatten elements
    process_list = list(src)
    i = 0
    while i < len(process_list):
        while isinstance(process_list[i], ltypes):
            if not process_list[i]:
                process_list.pop(i)
                i -= 1
                break
            else:
                process_list[i : i + 1] = process_list[i]
        i += 1

    # If input src data type is tuple, return tuple
    if not isinstance(process_list, ltype):
        return tuple(process_list)
    return process_list


def support_only_sql_alchemy_connection(command):
    """
    Throws a sql.exceptions.RuntimeError if connection is not SQLAlchemy
    """
    if ConnectionManager.current.is_dbapi_connection:
        raise exceptions.RuntimeError(
            f"{command} is only supported with SQLAlchemy "
            "connections, not with DBAPI connections"
        )


def fetch_sql_with_pagination(
    table, offset, n_rows, sort_column=None, sort_order=None
) -> tuple:
    """
    Returns next n_rows and columns from table starting at the offset

    Parameters
    ----------
    table : str
        Table name

    offset : int
        Specifies the number of rows to skip before
        it starts to return rows from the query expression.

    n_rows : int
        Number of rows to return.

    sort_column : str, default None
        Sort by column

    sort_order : 'DESC' or 'ASC', default None
        Order list
    """
    is_table_exists(table)

    order_by = "" if not sort_column else f"ORDER BY {sort_column} {sort_order}"

    query = f"""
    SELECT * FROM {table} {order_by}
    OFFSET {offset} ROWS FETCH NEXT {n_rows} ROWS ONLY"""

    rows = ConnectionManager.current.execute(query).fetchall()

    columns = ConnectionManager.current.raw_execute(
        f"SELECT * FROM {table} WHERE 1=0"
    ).keys()

    return rows, columns


def parse_sql_results_to_json(rows, columns) -> str:
    """
    Serializes sql rows to a JSON formatted ``str``
    """
    dicts = [dict(zip(list(columns), row)) for row in rows]
    rows_json = json.dumps(dicts, indent=4, sort_keys=True, default=str).replace(
        "null", '"None"'
    )

    return rows_json


def get_all_keys():
    """

    Returns
    -------
    All stored snippets in the current session
    """
    return list(store)


def get_key_dependents(key: str) -> list:
    """
    Function to find the stored snippets dependent on key
    Parameters
    ----------
    key : str, name of the table

    Returns
    -------
    list
        List of snippets dependent on key

    """
    deps = _get_dependents_for_key(store, key)
    return deps


def del_saved_key(key: str) -> str:
    """
    Deletes a stored snippet
    Parameters
    ----------
    key : str, name of the snippet to be deleted

    Returns
    -------
    list
        Remaining stored snippets
    """
    all_keys = get_all_keys()
    if key not in all_keys:
        raise exceptions.UsageError(f"No such saved snippet found : {key}")
    del store[key]
    return get_all_keys()


def show_deprecation_warning():
    """
    Raises CTE deprecation warning
    """
    warnings.warn(
        "CTE dependencies are now automatically inferred, "
        "you can omit the --with arguments. Using --with will "
        "raise an exception in the next major release so please remove it.",
        FutureWarning,
    )


def find_path_from_root(file_name):
    """
    Recursively finds an absolute path to file_name starting
    from current to root directory
    """
    current = Path().resolve()
    while not (current / file_name).exists():
        if current == current.parent:
            return None

        current = current.parent
    display.message(f"Found {file_name} from '{current}'")

    return str(Path(current, file_name))


def find_close_match_config(word, possibilities, n=3):
    """Finds closest matching configurations and displays message"""
    closest_matches = difflib.get_close_matches(word, possibilities, n=n)
    if not closest_matches:
        display.message_html(
            f"'{word}' is an invalid configuration. Please review our "
            "<a href='https://jupysql.ploomber.io/en/latest/api/configuration.html#options'>"  # noqa
            "configuration guideline</a>."
        )
    else:
        display.message(
            f"'{word}' is an invalid configuration. Did you mean "
            f"{pretty_print(closest_matches, last_delimiter='or')}?"
        )


def get_line_content_from_toml(file_path, line_number):
    """
    Locates a line that error occurs when loading a toml file
    and returns the line, key, and value
    """
    with open(file_path, "r") as file:
        lines = file.readlines()
        eline = lines[line_number - 1].strip()
        ekey, evalue = None, None
        if "=" in eline:
            ekey, evalue = map(str.strip, eline.split("="))
        return eline, ekey, evalue


@requires(["toml"])
def load_toml(file_path):
    """
    Returns toml file content in a dictionary format
    and raises error if it fails to load the toml file
    """
    try:
        with open(file_path, "r") as file:
            content = file.read()
            return toml.loads(content)
    except toml.TomlDecodeError as e:
        raise parse_toml_error(e, file_path)


def parse_toml_error(e, file_path):
    eline, ekey, evalue = get_line_content_from_toml(file_path, e.lineno)
    if "Duplicate keys!" in str(e):
        return exceptions.ConfigurationError(f"Duplicate key found : '{ekey}'")
    elif "Only all lowercase booleans" in str(e):
        return exceptions.ConfigurationError(
            f"Invalid value '{evalue}' in '{eline}'. "
            "Valid boolean values: true, false"
        )
    elif "invalid literal for int()" in str(e):
        return exceptions.ConfigurationError(
            f"Invalid value '{evalue}' in '{eline}'. "
            "To use str value, enclose it with ' or \"."
        )
    else:
        return e


def get_user_configs(file_path, section_names):
    """
    Returns saved configuration settings in a toml file from given file_path

    Parameters
    ----------
    file_path : str
        file path to a toml file
    section_names : list
        section names that contains the configuration settings
        (e.g., ["tool", "jupysql", "SqlMagic"])

    Returns
    -------
    dict
        saved configuration settings
    """
    data = load_toml(file_path)
    while section_names:
        section_to_find, sections_from_user = section_names.pop(0), data.keys()
        if section_to_find not in sections_from_user:
            close_match = difflib.get_close_matches(section_to_find, sections_from_user)
            if not close_match:
                return {}
            else:
                raise exceptions.ConfigurationError(
                    f"{pretty_print(close_match)} is an invalid section name. "
                    f"Did you mean '{section_to_find}'?"
                )
        data = data[section_to_find]
    return data


def get_default_configs(sql):
    """
    Returns a dictionary of SqlMagic configuration settings users can set
    with their default values.
    """
    default_configs = sql.trait_defaults()
    del default_configs["parent"]
    del default_configs["config"]
    return default_configs



import re
import operator
from functools import reduce
from io import StringIO
import html
from collections.abc import Iterable


import prettytable

from sql.column_guesser import ColumnGuesserMixin
from sql.run.csv import CSVWriter, CSVResultDescriptor
from sql.telemetry import telemetry
from sql.run.table import CustomPrettyTable


class ResultSet(ColumnGuesserMixin):
    """
    Results of a SQL query. Fetches rows lazily (only the necessary rows to show the
    preview based on the current configuration)
    """

    def __init__(self, sqlaproxy, config, statement=None, conn=None):
        self._closed = False
        self._config = config
        self._statement = statement
        self._sqlaproxy = sqlaproxy
        self._conn = conn
        self._dialect = conn._get_sqlglot_dialect()
        self._keys = None
        self._field_names = None
        self._results = []
        # https://peps.python.org/pep-0249/#description
        self._is_dbapi_results = hasattr(sqlaproxy, "description")

        # note that calling this will fetch the keys
        self._pretty_table = self._init_table()

        self._mark_fetching_as_done = False

        if self._config.autolimit == 1:
            # if autolimit is 1, we only want to fetch one row
            self.fetchmany(size=1)
            self._done_fetching()
        else:
            # in all other cases, 2 allows us to know if there are more rows
            # for example when creating a table, the results contains one row, in
            # such case, fetching 2 rows will tell us that there are no more rows
            # and can set the _mark_fetching_as_done flag to True
            self.fetchmany(size=2)

        self._finished_init = True

        if conn:
            conn._result_sets.append(self)

    @property
    def sqlaproxy(self):
        conn = self._conn

        # mssql with pyodbc does not support multiple open result sets, so we need
        # to close them all. when running this, we might've already closed the results
        # so we need to check for that and re-open the results if needed
        if conn.dialect == "mssql" and conn.driver == "pyodbc" and self._closed:
            self._conn._result_sets.close_all()
            self._sqlaproxy = self._conn.raw_execute(self._statement)
            self._sqlaproxy.fetchmany(size=len(self._results))
            self._conn._result_sets.append(self)

        # there is a problem when using duckdb + sqlalchemy: duckdb-engine doesn't
        # create separate cursors, so whenever we have >1 ResultSet, the old ones
        # become outdated and fetching their results will return the results from
        # the last ResultSet. To fix this, we have to re-issue the query
        is_last_result = self._conn._result_sets.is_last(self)

        is_duckdb_sqlalchemy = (
            self._dialect == "duckdb" and not self._conn.is_dbapi_connection
        )

        if (
            # skip this if we're initializing the object (we're running __init__)
            hasattr(self, "_finished_init")
            # this only applies to duckdb + sqlalchemy with outdated results
            and is_duckdb_sqlalchemy
            and not is_last_result
        ):
            self._sqlaproxy = self._conn.raw_execute(self._statement)
            self._sqlaproxy.fetchmany(size=len(self._results))

            # ensure we make his result set the last one
            self._conn._result_sets.append(self)

        return self._sqlaproxy

    def _extend_results(self, elements):
        """Store the DB fetched results into the internal list of results"""
        to_add = self._config.displaylimit - len(self._results)
        self._results.extend(elements)
        self._pretty_table.add_rows(elements[:to_add])

    def mark_fetching_as_done(self):
        self._mark_fetching_as_done = True
        # NOTE: don't close the connection here (self.sqlaproxy.close()),
        # because we need to keep it open for the next query

    def _done_fetching(self):
        return self._mark_fetching_as_done

    @property
    def field_names(self):
        if self._field_names is None:
            self._field_names = unduplicate_field_names(self.keys)

        return self._field_names

    @property
    def keys(self):
        """
        Return the keys of the results (the column names)
        """
        if self._keys is not None:
            return self._keys

        if not self._is_dbapi_results:
            try:
                self._keys = self.sqlaproxy.keys()
            # sqlite with sqlalchemy raises sqlalchemy.exc.ResourceClosedError,
            # psycopg2 raises psycopg2.ProgrammingError error when running a script
            # that doesn't return rows e.g, 'CREATE TABLE' but others don't
            # (e.g., duckdb), so here we catch all
            except Exception:
                self._keys = []
                return self._keys

        elif isinstance(self.sqlaproxy.description, Iterable):
            self._keys = [i[0] for i in self.sqlaproxy.description]
        else:
            self._keys = []

        return self._keys

    def _repr_html_(self):
        self.fetch_for_repr_if_needed()

        _cell_with_spaces_pattern = re.compile(r"(<td>)( {2,})")

        result = self._pretty_table.get_html_string()

        HTML = (
            "%s\n<span style='font-style:italic;font-size:11px'>"
            "<code>ResultSet</code> : to convert to pandas, call <a href="
            "'https://jupysql.ploomber.io/en/latest/integrations/pandas.html'>"
            "<code>.DataFrame()</code></a> or to polars, call <a href="
            "'https://jupysql.ploomber.io/en/latest/integrations/polars.html'>"
            "<code>.PolarsDataFrame()</code></a></span><br>"
        )
        result = HTML % (result)

        # to create clickable links
        result = html.unescape(result)
        result = _cell_with_spaces_pattern.sub(_nonbreaking_spaces, result)

        if self._config.displaylimit != 0 and not self._done_fetching():
            HTML = (
                '%s\n<span style="font-style:italic;text-align:center;">'
                "Truncated to displaylimit of %d</span>"
                "<br>"
                '<span style="font-style:italic;text-align:center;">'
                "If you want to see more, please visit "
                '<a href="https://jupysql.ploomber.io/en/latest/api/configuration.html#displaylimit">displaylimit</a>'  # noqa: E501
                " configuration</span>"
            )
            result = HTML % (result, self._config.displaylimit)
        return result

    def __len__(self):
        self.fetchall()

        return len(self._results)

    def __iter__(self):
        self.fetchall()

        for result in self._results:
            yield result

    def __str__(self):
        self.fetch_for_repr_if_needed()
        return str(self._pretty_table)

    def __repr__(self) -> str:
        return str(self)

    def __eq__(self, another: object) -> bool:
        return self._results == another

    def __getitem__(self, key):
        """
        Access by integer (row position within result set)
        or by string (value of leftmost column)
        """
        try:
            return self._results[key]
        except TypeError:
            result = [row for row in self if row[0] == key]
            if not result:
                raise KeyError(key)
            if len(result) > 1:
                raise KeyError('%d results for "%s"' % (len(result), key))
            return result[0]

    def __getattr__(self, attr):
        err_msg = (
            f"'{attr}' is not a valid operation, you can convert this "
            "into a pandas data frame by calling '.DataFrame()' or a "
            "polars data frame by calling '.PolarsDataFrame()'"
        )
        raise AttributeError(err_msg)

    def dict(self):
        """Returns a single dict built from the result set

        Keys are column names; values are a tuple"""
        return dict(zip(self.keys, zip(*self)))

    def dicts(self):
        "Iterator yielding a dict for each row"
        for row in self:
            yield dict(zip(self.keys, row))

    @telemetry.log_call("data-frame", payload=True)
    def DataFrame(self, payload):
        """Returns a Pandas DataFrame instance built from the result set."""
        payload["connection_info"] = self._conn._get_database_information()
        import pandas as pd

        return _convert_to_data_frame(self, "df", pd.DataFrame)

    @telemetry.log_call("polars-data-frame")
    def PolarsDataFrame(self, **polars_dataframe_kwargs):
        """Returns a Polars DataFrame instance built from the result set."""
        import polars as pl

        polars_dataframe_kwargs["schema"] = self.keys
        return _convert_to_data_frame(self, "pl", pl.DataFrame, polars_dataframe_kwargs)

    @telemetry.log_call("pie")
    def pie(self, key_word_sep=" ", title=None, **kwargs):
        """Generates a pylab pie chart from the result set.

        ``matplotlib`` must be installed, and in an
        IPython Notebook, inlining must be on::

            %%matplotlib inline

        Values (pie slice sizes) are taken from the
        rightmost column (numerical values required).
        All other columns are used to label the pie slices.

        Parameters
        ----------
        key_word_sep: string used to separate column values
                      from each other in pie labels
        title: Plot title, defaults to name of value column

        Any additional keyword arguments will be passed
        through to ``matplotlib.pylab.pie``.
        """
        self.guess_pie_columns(xlabel_sep=key_word_sep)
        import matplotlib.pylab as plt

        ax = plt.gca()

        ax.pie(self.ys[0], labels=self.xlabels, **kwargs)
        ax.set_title(title or self.ys[0].name)
        return ax

    @telemetry.log_call("plot")
    def plot(self, title=None, **kwargs):
        """Generates a pylab plot from the result set.

        ``matplotlib`` must be installed, and in an
        IPython Notebook, inlining must be on::

            %%matplotlib inline

        The first and last columns are taken as the X and Y
        values.  Any columns between are ignored.

        Parameters
        ----------
        title: Plot title, defaults to names of Y value columns

        Any additional keyword arguments will be passed
        through to ``matplotlib.pylab.plot``.
        """
        import matplotlib.pylab as plt

        self.guess_plot_columns()
        self.x = self.x or range(len(self.ys[0]))

        ax = plt.gca()

        coords = reduce(operator.add, [(self.x, y) for y in self.ys])
        ax.plot(*coords, **kwargs)

        if hasattr(self.x, "name"):
            ax.set_xlabel(self.x.name)

        ylabel = ", ".join(y.name for y in self.ys)

        ax.set_title(title or ylabel)
        ax.set_ylabel(ylabel)

        return ax

    @telemetry.log_call("bar")
    def bar(self, key_word_sep=" ", title=None, **kwargs):
        """Generates a pylab bar plot from the result set.

        ``matplotlib`` must be installed, and in an
        IPython Notebook, inlining must be on::

            %%matplotlib inline

        The last quantitative column is taken as the Y values;
        all other columns are combined to label the X axis.

        Parameters
        ----------
        title: Plot title, defaults to names of Y value columns
        key_word_sep: string used to separate column values
                      from each other in labels

        Any additional keyword arguments will be passed
        through to ``matplotlib.pylab.bar``.
        """
        import matplotlib.pylab as plt

        ax = plt.gca()

        self.guess_pie_columns(xlabel_sep=key_word_sep)
        ax.bar(range(len(self.ys[0])), self.ys[0], **kwargs)

        if self.xlabels:
            ax.set_xticks(range(len(self.xlabels)), self.xlabels, rotation=45)

        ax.set_xlabel(self.xlabel)
        ax.set_ylabel(self.ys[0].name)
        return ax

    @telemetry.log_call("generate-csv")
    def csv(self, filename=None, **format_params):
        """Generate results in comma-separated form.  Write to ``filename`` if given.
        Any other parameters will be passed on to csv.writer."""
        if filename:
            encoding = format_params.get("encoding", "utf-8")
            outfile = open(filename, "w", newline="", encoding=encoding)
        else:
            outfile = StringIO()

        writer = CSVWriter(outfile, **format_params)
        writer.writerow(self.field_names)
        for row in self:
            writer.writerow(row)
        if filename:
            outfile.close()
            return CSVResultDescriptor(filename)
        else:
            return outfile.getvalue()

    def fetchmany(self, size):
        """Fetch n results and add it to the results"""
        if not self._done_fetching():
            try:
                returned = self.sqlaproxy.fetchmany(size=size)
            # sqlite with sqlalchemy raises sqlalchemy.exc.ResourceClosedError,
            # psycopg2 raises psycopg2.ProgrammingError error when running a script
            # that doesn't return rows e.g, 'CREATE TABLE' but others don't
            # (e.g., duckdb), so here we catch all
            except Exception:
                self.mark_fetching_as_done()
                return

            self._extend_results(returned)

            if len(returned) < size:
                self.mark_fetching_as_done()

            if (
                self._config.autolimit is not None
                and self._config.autolimit != 0
                and len(self._results) >= self._config.autolimit
            ):
                self.mark_fetching_as_done()

    def fetch_for_repr_if_needed(self):
        if self._config.displaylimit == 0:
            self.fetchall()

        missing = self._config.displaylimit - len(self._results)

        if missing > 0:
            self.fetchmany(missing)

    def fetchall(self):
        if not self._done_fetching():
            self._extend_results(self.sqlaproxy.fetchall())
            self.mark_fetching_as_done()

    def _init_table(self):
        pretty = CustomPrettyTable(self.field_names)

        if isinstance(self._config.style, str):
            _style = prettytable.__dict__[self._config.style.upper()]
            pretty.set_style(_style)

        return pretty

    def close(self):
        self._sqlaproxy.close()
        self._closed = True


def unduplicate_field_names(field_names):
    """Append a number to duplicate field names to make them unique."""
    res = []
    for k in field_names:
        if k in res:
            i = 1
            while k + "_" + str(i) in res:
                i += 1
            k += "_" + str(i)
        res.append(k)
    return res


def _convert_to_data_frame(
    result_set, converter_name, constructor, constructor_kwargs=None
):
    """
    Convert the result set to a pandas DataFrame, using native DuckDB methods if
    possible
    """
    constructor_kwargs = constructor_kwargs or {}

    # maybe create accessors in the connection objects?
    if result_set._conn.is_dbapi_connection:
        native_connection = result_set.sqlaproxy
    else:
        native_connection = result_set._conn._connection.connection

    has_converter_method = hasattr(native_connection, converter_name)

    # native duckdb connection
    if has_converter_method:
        # we need to re-execute the statement because if we fetched some rows
        # already, .df() will return None. But only if it's a select statement
        # otherwise we might end up re-execute INSERT INTO or CREATE TABLE
        # statements
        is_select = _statement_is_select(result_set._statement)

        if is_select:
            native_connection.execute(result_set._statement)

        return getattr(native_connection, converter_name)()
    else:
        if converter_name == "df":
            constructor_kwargs["columns"] = result_set.keys

        frame = constructor(
            (tuple(row) for row in result_set),
            **constructor_kwargs,
        )

        return frame


def _nonbreaking_spaces(match_obj):
    """
    Make spaces visible in HTML by replacing all `` `` with ``&nbsp;``

    Call with a ``re`` match object.  Retain group 1, replace group 2
    with nonbreaking spaces.
    """
    spaces = "&nbsp;" * len(match_obj.group(2))
    return "%s%s" % (match_obj.group(1), spaces)


def _statement_is_select(statement):
    statement_ = statement.lower().strip()
    # duckdb also allows FROM without SELECT
    return (
        statement_.startswith("select")
        or statement_.startswith("from")
        or statement_.startswith("with")
    )



from sql import inspect
from sql.cmd.cmd_utils import CmdParser


def tables(others):
    """
    Implementation of `%sqlcmd tables`

    This function takes in a string containing command line arguments,
    parses them to extract the schema name, and returns a list of table names
    present in the specified schema or in the default schema if none is specified.

    Parameters
    ----------
    others : str,
            A string containing the command line arguments.

    Returns
    -------
    table_names: list
        list of tables in the schema

    """
    parser = CmdParser()

    parser.add_argument("-s", "--schema", type=str, help="Schema name", required=False)

    args = parser.parse_args(others)

    return inspect.get_table_names(schema=args.schema)



from sql import exceptions
import sql.connection
from sqlglot import select, condition
from prettytable import PrettyTable
from sql.cmd.cmd_utils import CmdParser


def return_test_results(args, conn, query):
    columns = []

    try:
        column_data = conn.execute(query).cursor.description
        res = conn.execute(query).fetchall()
        for column in column_data:
            columns.append(column[0])
        res = [columns, *res]
        return res
    except Exception as e:
        if "column" in str(e):
            raise exceptions.UsageError(
                f"Referenced column '{args.column}' not found!"
            ) from e


def run_each_individually(args, conn):
    base_query = select("*").from_(args.table)

    storage = {}

    if args.greater:
        where = condition(args.column + "<=" + args.greater)
        current_query = base_query.where(where).sql()

        res = return_test_results(args, conn, query=current_query)

        if res is not None:
            storage["greater"] = res
    if args.greater_or_equal:
        where = condition(args.column + "<" + args.greater_or_equal)

        current_query = base_query.where(where).sql()

        res = return_test_results(args, conn, query=current_query)

        if res is not None:
            storage["greater_or_equal"] = res

    if args.less_than_or_equal:
        where = condition(args.column + ">" + args.less_than_or_equal)
        current_query = base_query.where(where).sql()

        res = return_test_results(args, conn, query=current_query)

        if res is not None:
            storage["less_than_or_equal"] = res
    if args.less_than:
        where = condition(args.column + ">=" + args.less_than)
        current_query = base_query.where(where).sql()

        res = return_test_results(args, conn, query=current_query)

        if res is not None:
            storage["less_than"] = res
    if args.no_nulls:
        where = condition("{} is NULL".format(args.column))
        current_query = base_query.where(where).sql()

        res = return_test_results(args, conn, query=current_query)

        if res is not None:
            storage["null"] = res

    return storage


def test(others):
    """
    Implementation of `%sqlcmd test`

    This function takes in a string containing command line arguments,
    parses them to extract the table name, column name, and conditions
    to return if those conditions are satisfied in that table

    Parameters
    ----------
    others : str,
            A string containing the command line arguments.

    Returns
    -------
    result: bool
        Result of the test

    table: PrettyTable
        table with rows because of which the test fails


    """
    parser = CmdParser()

    parser.add_argument("-t", "--table", type=str, help="Table name", required=True)
    parser.add_argument("-c", "--column", type=str, help="Column name", required=False)
    parser.add_argument(
        "-g",
        "--greater",
        type=str,
        help="Greater than a certain number.",
        required=False,
    )
    parser.add_argument(
        "-goe",
        "--greater-or-equal",
        type=str,
        help="Greater or equal than a certain number.",
        required=False,
    )
    parser.add_argument(
        "-l",
        "--less-than",
        type=str,
        help="Less than a certain number.",
        required=False,
    )
    parser.add_argument(
        "-loe",
        "--less-than-or-equal",
        type=str,
        help="Less than or equal to a certain number.",
        required=False,
    )
    parser.add_argument(
        "-nn",
        "--no-nulls",
        help="Returns rows in specified column that are not null.",
        action="store_true",
    )

    args = parser.parse_args(others)

    COMPARATOR_ARGS = [
        args.greater,
        args.greater_or_equal,
        args.less_than,
        args.less_than_or_equal,
    ]

    if args.table and not any(COMPARATOR_ARGS):
        raise exceptions.UsageError("Please use a valid comparator.")

    if args.table and any(COMPARATOR_ARGS) and not args.column:
        raise exceptions.UsageError("Please pass a column to test.")

    if args.greater and args.greater_or_equal:
        return exceptions.UsageError(
            "You cannot use both greater and greater "
            "than or equal to arguments at the same time."
        )
    elif args.less_than and args.less_than_or_equal:
        return exceptions.UsageError(
            "You cannot use both less and less than "
            "or equal to arguments at the same time."
        )

    conn = sql.connection.ConnectionManager.current
    result_dict = run_each_individually(args, conn)

    if any(len(rows) > 1 for rows in list(result_dict.values())):
        for comparator, rows in result_dict.items():
            if len(rows) > 1:
                print(f"\n{comparator}:\n")
                _pretty = PrettyTable()
                _pretty.field_names = rows[0]
                for row in rows[1:]:
                    _pretty.add_row(row)
                print(_pretty)
        raise exceptions.UsageError(
            "The above values do not match your test requirements."
        )
    else:
        return True



import sqlparse

from sql import exceptions, display
from sql.run.resultset import ResultSet
from sql.run.pgspecial import handle_postgres_special


# TODO: conn also has access to config, we should clean this up to provide a clean
# way to access the config
def run_statements(conn, sql, config, parameters=None):
    """
    Run a SQL query (supports running multiple SQL statements) with the given
    connection. This is the function that's called when executing SQL magic.

    Parameters
    ----------
    conn : sql.connection.AbstractConnection
        The connection to use

    sql : str
        SQL query to execution

    config
        Configuration object
    """
    if not sql.strip():
        return "Connected: %s" % conn.name

    for statement in sqlparse.split(sql):
        first_word = sql.strip().split()[0].lower()

        if first_word == "begin":
            raise exceptions.RuntimeError("JupySQL does not support transactions")

        # postgres metacommand
        if first_word.startswith("\\") and is_postgres_or_redshift(conn.dialect):
            result = handle_postgres_special(conn, statement)

        # regular query
        else:
            result = conn.raw_execute(statement, parameters=parameters)

            if config.feedback and hasattr(result, "rowcount") and result.rowcount > 0:
                display.message_success(f"{result.rowcount} rows affected.")

    result_set = ResultSet(result, config, statement, conn)
    return select_df_type(result_set, config)


def is_postgres_or_redshift(dialect):
    """Checks if dialect is postgres or redshift"""
    return "postgres" in str(dialect) or "redshift" in str(dialect)


def select_df_type(resultset, config):
    """
    Converts the input resultset to either a Pandas DataFrame
    or Polars DataFrame based on the config settings.
    """
    if config.autopandas:
        return resultset.DataFrame()
    elif config.autopolars:
        return resultset.PolarsDataFrame(**config.polars_dataframe_kwargs)
    else:
        return resultset



"""
A module to display confirmation messages and contextual information to the user
"""
import html

from prettytable import PrettyTable
from IPython.display import display, HTML


class Table:
    """Provides a txt and html representation of tabular data"""

    TITLE = ""

    def __init__(self, headers, rows) -> None:
        self._headers = headers
        self._rows = rows
        self._table = PrettyTable()
        self._table.field_names = headers

        for row in rows:
            self._table.add_row(row)

        self._table_html = self._table.get_html_string()
        self._table_txt = self._table.get_string()

    def __repr__(self) -> str:
        return self.TITLE + "\n" + self._table_txt

    def _repr_html_(self) -> str:
        return self.TITLE + "\n" + self._table_html


class ConnectionsTable(Table):
    TITLE = "Active connections:"

    def __init__(self, headers, rows_maps) -> None:
        def get_values(d):
            d = {k: v for k, v in d.items() if k not in {"connection", "key"}}
            return list(d.values())

        rows = [get_values(r) for r in rows_maps]

        self._mapping = {}

        for row in rows_maps:
            self._mapping[row["key"]] = row["connection"]

        super().__init__(headers=headers, rows=rows)

    def __getitem__(self, key: str):
        """
        This method is provided for backwards compatibility. Before
        creating ConnectionsTable, `%sql --connections` returned a dictionary,
        hence users could retrieve connections using __getitem__. Note that this
        was undocumented so we might decide to remove it in the future.
        """
        return self._mapping[key]

    def __iter__(self):
        """Also provided for backwards compatibility"""
        for key in self._mapping:
            yield key

    def __len__(self):
        """Also provided for backwards compatibility"""
        return len(self._mapping)


class Message:
    """Message for the user"""

    def __init__(self, message, style=None) -> None:
        self._message = message
        self._message_html = html.escape(message)
        self._style = "" or style

    def _repr_html_(self):
        return f'<span style="{self._style}">{self._message_html}</span>'

    def __repr__(self) -> str:
        return self._message


def message(message):
    """Display a generic message"""
    display(Message(message))


def message_success(message):
    """Display a success message"""
    display(Message(message, style="color: green"))


def message_html(message):
    """Display a message as HTML"""
    display(HTML(str(Message(message))))


def table(headers, rows):
    """Display a table"""
    display(Table(headers, rows))



from sql import inspect
from sql.util import sanitize_identifier
from sql.cmd.cmd_utils import CmdParser


def columns(others):
    """
    Implementation of `%sqlcmd columns`
    This function takes in a string containing command line arguments,
    parses them to extract the name of the table and the schema, and returns
    a list of columns for the specified table.

    Parameters
    ----------
    others : str,
        A string containing the command line arguments.

    Returns
    -------
    columns: list
        information of the columns in the specified table
    """
    parser = CmdParser()

    parser.add_argument("-t", "--table", type=str, help="Table name", required=True)
    parser.add_argument("-s", "--schema", type=str, help="Schema name", required=False)

    args = parser.parse_args(others)
    return inspect.get_columns(name=sanitize_identifier(args.table), schema=args.schema)



from sql.ggplot.ggplot import ggplot
from sql.ggplot.aes import aes
from sql.ggplot.geom import geom_boxplot, geom_histogram
from sql.ggplot.facet_wrap import facet_wrap


__all__ = ["ggplot", "aes", "geom_boxplot", "geom_histogram", "facet_wrap"]
